from cicinlang.nodes import *
from cicinlang.utils import Constants, Token
from cicinlang.errors import SyntaxError

class Parser:
    def __init__( self, tokens ):
        self.tokens = tokens
        self.idx = -1
        self.advance()
    
    def advance( self ):
        self.idx += 1
        self.cur_tok = self.tokens[self.idx] if self.idx < len( self.tokens ) else None 
    
    def parse( self ):
        exprs = []

        while self.cur_tok.type != Constants.TT_EOF:
            head, error = self.expr()

            if error: return None, error 

            node_type = type( head ).__name__

            semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( head.var_value_node ).__name__ != 'FunctionDefNode' )

            if semicolon_required:
                if self.cur_tok.type != Constants.TT_SEMICOLON:
                    return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                
                self.advance()

            exprs.append( head )
        
        return exprs, None 

    def bin_op_term( self, parse_func, tok_val_func ):
        left, error = parse_func()

        if error: return None, error 

        while tok_val_func( self.cur_tok ):
            tok = self.cur_tok 
            self.advance()
            right, error = parse_func()

            if error: return None, error 
            
            left = BinOpNode( left, tok, right )
        
        return left, None 
    
    def expr( self, inside_func=False ):
        expr_head, error = self.void_expr( inside_func ) if self.cur_tok.value in [ 'for', 'if', 'print', 'return' ] else self.value_expr()

        if error: return None, error 

        return expr_head, None 

    def value_expr( self ):
        if self.cur_tok.value == 'var': 
            self.advance()

            name_tok = self.cur_tok 

            if name_tok.type != Constants.TT_IDENTIFIER:
                return None, SyntaxError( "Expected Identifier", self.cur_tok.start_pos, self.cur_tok.end_pos )
            
            self.advance()

            if self.cur_tok.type != Constants.TT_EQ:
                return None, SyntaxError( "Expected '='", self.cur_tok.start_pos, self.cur_tok.end_pos )
            
            self.advance()

            value_node, error = self.value_expr()

            if error: return None, error 

            return VarAssignNode( name_tok, value_node, Constants.AT_NEW ), None

        head, error = self.or_chain()

        if error: return None, error 

        if type( head ).__name__ == 'VarAccessNode' and self.cur_tok.type == Constants.TT_EQ:
            name_tok = head.var_name_token

            self.advance()

            value_node, error = self.value_expr()

            if error: return None, error 

            return VarAssignNode( name_tok, value_node, Constants.AT_OLD ), None
        
        return head, None 

    def void_expr( self, inside_func=False ):
        if self.cur_tok.value == 'for':
            for_head, error = self.for_loop( inside_func )

            if error: return None, error 

            return for_head, None 
        
        elif self.cur_tok.value == 'if':
            if_head, error = self.if_stmt( inside_func )

            if error: return None, error 

            return if_head, None 
        
        elif self.cur_tok.value == 'print':
            print_head, error = self.print_stmt()

            if error: return None, error 

            return print_head, None 
        
        elif self.cur_tok.value == 'return':
            if not inside_func: 
                return None, SyntaxError( "Return statements only allowed as standalone statements inside functions", self.cur_tok.start_pos, self.cur_tok.end_pos )
            
            self.advance()

            head, error = self.value_expr()

            if error: return None, error 

            return ReturnNode( head ), None 

    def or_chain( self ):
        return self.bin_op_term( self.or_operand, lambda x : x.value == 'or' )
    
    def or_operand( self ):
        return self.bin_op_term( self.and_operand, lambda x : x.value == 'and' )
    
    def and_operand( self ):
        return self.bin_op_term( self.ee_operand, lambda x : x.type in [ Constants.TT_EE, Constants.TT_NE ] )
    
    def ee_operand( self ):
        return self.bin_op_term( self.gt_operand, lambda x : x.type in [ Constants.TT_GT, Constants.TT_GTE ] )
    
    def gt_operand( self ):
        return self.bin_op_term( self.lt_operand, lambda x : x.type in [ Constants.TT_LT, Constants.TT_LTE ] )
    
    def lt_operand( self ):
        return self.bin_op_term( self.term, lambda x : x.type in [ Constants.TT_PLUS, Constants.TT_MINUS ] )
    
    def term( self ):
        return self.bin_op_term( self.factor, lambda x : x.type in [ Constants.TT_MUL, Constants.TT_DIV ] )
    
    def factor( self ):
        atoms = []

        head, error = self.atom()

        if error: return None, error 

        if type( head ).__name__ == 'FunctionDefNode': return head, None

        atoms.append( head )

        while self.cur_tok.type == Constants.TT_EXP:
            self.advance()

            head, error = self.atom()

            if error: return None, error 

            atoms.append( head )
        
        while len( atoms ) > 1:
            right = atoms.pop()
            left = atoms.pop()

            atoms.append( BinOpNode( left, Token( Constants.TT_EXP ), right ) )

        return atoms[0], None

    def atom( self ):
        if self.cur_tok.type in [ Constants.TT_INT, Constants.TT_FLOAT ]:
            tok = self.cur_tok 
            self.advance()

            return NumberNode( tok ), None 
        
        elif self.cur_tok.type == Constants.TT_STRING:
            tok = self.cur_tok
            self.advance()

            return StringNode( tok ), None

        elif self.cur_tok.type == Constants.TT_IDENTIFIER:
            tok = self.cur_tok 
            self.advance()

            if self.cur_tok.type != Constants.TT_LPAREN:
                return VarAccessNode( tok ), None 
            
            self.advance()

            func_arg_nodes_list = [] 

            while self.cur_tok.type != Constants.TT_RPAREN:
                head, error = self.value_expr()

                if error: return None, error 

                func_arg_nodes_list.append( head )

                if self.cur_tok.type == Constants.TT_COMMA:
                    self.advance()
                
                elif self.cur_tok.type != Constants.TT_RPAREN:
                    return None, SyntaxError( "Expected ',' or ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
            
            self.advance()

            return FunctionCallNode( tok, func_arg_nodes_list ), None
        
        elif self.cur_tok.type == Constants.TT_LPAREN:
            self.advance()

            if self.cur_tok.type == Constants.TT_RPAREN:
                func_arg_toks_list = []
                func_body_nodes_list = []

                self.advance()

                if self.cur_tok.type != Constants.TT_LBRACE:
                    return None, SyntaxError( "Expected '{'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                
                lbrace_tok = self.cur_tok 

                self.advance()

                while self.cur_tok.type != Constants.TT_RBRACE:
                    head, error = self.expr( True )

                    if error: return None, error 

                    node_type = type( head ).__name__

                    semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( head.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( head.node ).__name__ != 'FunctionDefNode' )

                    if semicolon_required:
                        if self.cur_tok.type != Constants.TT_SEMICOLON:
                            return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                        
                        self.advance()
                    
                    func_body_nodes_list.append( head )

                if len( func_body_nodes_list ) == 0:
                    return None, SyntaxError( "Functions with empty bodies are not supported", lbrace_tok.start_pos, self.cur_tok.end_pos )
                
                self.advance()

                return FunctionDefNode( func_arg_toks_list, func_body_nodes_list ), None 

            head, error = self.value_expr()

            if error: return None, error 

            nt = type( head ).__name__

            if nt == 'VarAccessNode' and self.cur_tok.type == Constants.TT_COMMA:
                func_arg_toks_list = [ head.var_name_token ]

                self.advance()

                while self.cur_tok.type != Constants.TT_RPAREN:
                    head, error = self.value_expr()

                    if error: return None, error 

                    if type( head ).__name__ != 'VarAccessNode':
                        return None, SyntaxError( "Expected identifier", head.start_pos, head.end_pos )
                    
                    func_arg_toks_list.append( head.var_name_token )
                
                self.advance()

                if self.cur_tok.type != Constants.TT_LBRACE:
                    return None, SyntaxError( "Expected '{'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                
                self.advance()

                func_body_nodes_list = []

                while self.cur_tok.type != Constants.TT_RBRACE:
                    head, error = self.expr( True )

                    if error: return None, error

                    node_type = type( head ).__name__

                    semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( head.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( head.node ).__name__ != 'FunctionDefNode' )

                    if semicolon_required:
                        if self.cur_tok.type != Constants.TT_SEMICOLON:
                            return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                        
                        self.advance()
                    
                    func_body_nodes_list.append( head )

                if len( func_body_nodes_list ) == 0:
                    return None, SyntaxError( "Functions with empty bodies are not supported", lbrace_tok.start_pos, self.cur_tok.end_pos )

                self.advance()

                return FunctionDefNode( func_arg_toks_list, func_body_nodes_list ), None

            if self.cur_tok.type != Constants.TT_RPAREN:
                return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
            
            self.advance()

            func_arg_toks_list = [ head.var_name_token ]

            if nt != 'VarAccessNode' or self.cur_tok.type != Constants.TT_LBRACE:
                return head, None 
            
            self.advance()

            func_body_nodes_list = []

            while self.cur_tok.type != Constants.TT_RBRACE:
                head, error = self.expr( True )
                
                if error: return None, error

                node_type = type( head ).__name__

                semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( head.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( head.node ).__name__ != 'FunctionDefNode' )

                if semicolon_required:
                    if self.cur_tok.type != Constants.TT_SEMICOLON:
                        return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                    
                    self.advance()
                
                func_body_nodes_list.append( head )

            self.advance()

            return FunctionDefNode( func_arg_toks_list, func_body_nodes_list ), None
        
        elif self.cur_tok.type in [ Constants.TT_PLUS, Constants.TT_MINUS ] or self.cur_tok.value == 'not':
            tok = self.cur_tok

            self.advance()

            node, error = self.atom()

            if error: return None, error 

            return UnOpNode( tok, node ), None 

        elif self.cur_tok.value in [ 'input_str', 'input_num' ]:
            input_head, error = self.input_stmt()

            if error: return None, error 

            return input_head, None

        elif self.cur_tok.value == 'str':
            str_head, error = self.stringify_stmt()

            if error: return None, error 

            return str_head, None

        else:
            return None, SyntaxError( "Expected int, float, identifier, '(', '+', '-' or 'not'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
    def if_stmt( self, inside_func=False ):
        self.advance()

        if self.cur_tok.type != Constants.TT_LPAREN:
            return None, SyntaxError( "Expected '('", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if_condition_node, error = self.value_expr()

        if error: return None, error 

        if self.cur_tok.type != Constants.TT_RPAREN:
            return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if self.cur_tok.type != Constants.TT_LBRACE:
            return None, SyntaxError( "Expected '{'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if_body_node_list = []

        while self.cur_tok.type != Constants.TT_RBRACE:
            if_body_node, error = self.expr( inside_func )

            if error: return None, error 

            node_type = type( if_body_node ).__name__

            semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( if_body_node.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( if_body_node.node ).__name__ != 'FunctionDefNode' )

            if semicolon_required:
                if self.cur_tok.type != Constants.TT_SEMICOLON:
                    return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                
                self.advance()

            if_body_node_list.append( if_body_node )
        
        self.advance()

        elif_cond_node_list = []
        elif_body_node_lists = []

        while self.cur_tok.value == 'elif':
            self.advance()

            if self.cur_tok.type != Constants.TT_LPAREN:
                return None, SyntaxError( "Expected '('", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
            self.advance()

            elif_cond_node, error = self.value_expr()

            if error: return None, error 

            elif_cond_node_list.append( elif_cond_node )

            if self.cur_tok.type != Constants.TT_RPAREN:
                return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
            self.advance()

            if self.cur_tok.type != Constants.TT_LBRACE:
                return None, SyntaxError( "Expected '{'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
            self.advance()

            elif_body_node_list = []

            while self.cur_tok.type != Constants.TT_RBRACE:
                elif_body_node, error = self.expr( inside_func )

                if error: return None, error 

                node_type = type( elif_body_node ).__name__

                semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( elif_body_node.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( elif_body_node.node ).__name__ != 'FunctionDefNode' )

                if semicolon_required:
                    if self.cur_tok.type != Constants.TT_SEMICOLON:
                        return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                    
                    self.advance()

                elif_body_node_list.append( elif_body_node )
            
            self.advance()

            elif_body_node_lists.append( elif_body_node_list )

        if self.cur_tok.value != 'else':
            return None, SyntaxError( "Expected 'else'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if self.cur_tok.type != Constants.TT_LBRACE:
            return None, SyntaxError( "Expected '{'", self.cur_tok.start_pos, self.cur_tok.end_pos )

        self.advance()

        else_body_node_list = []

        while self.cur_tok.type != Constants.TT_RBRACE:
            else_body_node, error = self.expr( inside_func )

            if error: return None, error 

            node_type = type( else_body_node ).__name__

            semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( else_body_node.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( else_body_node.node ).__name__ != 'FunctionDefNode' )

            if semicolon_required:
                if self.cur_tok.type != Constants.TT_SEMICOLON:
                    return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                
                self.advance()

            else_body_node_list.append( else_body_node )
        
        self.advance()

        return IfNode( if_condition_node, if_body_node_list, elif_cond_node_list, elif_body_node_lists, else_body_node_list ), None

    def for_loop( self, inside_func=False ):
        init_node = None 

        self.advance()

        if self.cur_tok.type != Constants.TT_LPAREN:
            return None, SyntaxError( "Expected '('", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if self.cur_tok.type != Constants.TT_SEMICOLON:
            init_node, error = self.expr()

            if error: return None, error 

        if self.cur_tok.type != Constants.TT_SEMICOLON:
            return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        cond_node, error = self.value_expr()

        if error: return None, error 

        if self.cur_tok.type != Constants.TT_SEMICOLON:
            return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        update_node, error = self.expr()

        if error: return None, error 

        if self.cur_tok.type != Constants.TT_RPAREN:
            return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if self.cur_tok.type != Constants.TT_LBRACE:
            return None, SyntaxError( "Expected '{'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        body_node_list = []

        while self.cur_tok.type != Constants.TT_RBRACE:
            head, error = self.expr( inside_func )

            if error: return None, error 

            node_type = type( head ).__name__

            semicolon_required = node_type not in [ "ForNode", "IfNode", "FunctionDefNode" ] and ( node_type != 'VarAssignNode' or type( head.var_value_node ).__name__ != 'FunctionDefNode' ) and ( node_type != 'ReturnNode' or type( head.node ).__name__ != 'FunctionDefNode' )

            if semicolon_required:
                if self.cur_tok.type != Constants.TT_SEMICOLON:
                    return None, SyntaxError( "Expected ';'", self.cur_tok.start_pos, self.cur_tok.end_pos )
                
                self.advance()

            body_node_list.append( head )
        
        self.advance()

        return ForNode( init_node, cond_node, update_node, body_node_list ), None

    def print_stmt( self ):
        self.advance()

        if self.cur_tok.type != Constants.TT_LPAREN:
            return None, SyntaxError( "Expected '('", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        head, error = self.value_expr()

        if error: return None, error 

        if self.cur_tok.type != Constants.TT_RPAREN:
            return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        return PrintNode( head ), None
    
    def input_stmt( self ):
        inp_type = 'str' if self.cur_tok.value == 'input_str' else 'num'
        self.advance()

        if self.cur_tok.type != Constants.TT_LPAREN:
            return None, SyntaxError( "Expected '('", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        if self.cur_tok.type == Constants.TT_RPAREN:
            self.advance()
            return InputNode( None, inp_type ), None

        head, error = self.value_expr()

        if error: return None, error 

        if self.cur_tok.type != Constants.TT_RPAREN:
            return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        return InputNode( head, inp_type ), None
    
    def stringify_stmt( self ): 
        self.advance()

        if self.cur_tok.type != Constants.TT_LPAREN:
            return None, SyntaxError( "Expected '('", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        head, error = self.value_expr()

        if error: return None, error 

        if self.cur_tok.type != Constants.TT_RPAREN:
            return None, SyntaxError( "Expected ')'", self.cur_tok.start_pos, self.cur_tok.end_pos )
        
        self.advance()

        return StringifyNode( head ), None



class Exception:
    def __init__( self, name, desc, start_pos, end_pos=None ):
        self.name = name 
        self.desc = desc
        self.start_pos = start_pos 

        if end_pos == None and start_pos != None:
            end_pos = start_pos.copy()
            end_pos.advance()
            self.end_pos = end_pos
        
        else:
            self.end_pos = end_pos 
    
    def __repr__( self ):
        res = f'{self.name} : {self.desc}\n'
        if self.start_pos != None:
            res += f'Line {self.start_pos.ln + 1}, Col {self.start_pos.coln + 1}'

        return res
    

class Error:
    def __init__( self, name, desc, start_pos, end_pos=None ):
        self.name = name 
        self.desc = desc
        self.start_pos = start_pos 
        self.end_pos = end_pos 
    
    def __repr__( self ):
        res = f'{self.name} : {self.desc}\n'
        res += f'Line {self.start_pos.ln + 1}, Col {self.start_pos.coln + 1}'

        return res


class IllegalCharException( Exception ):
    def __init__( self, desc, start_pos, end_pos=None ):
        super().__init__( 'Illegal Character', desc, start_pos, end_pos )


class RuntimeException( Exception ):
    def __init__( self, desc, start_pos, end_pos=None ):
        super().__init__( 'Runtime Exception', desc, start_pos, end_pos )


class SyntaxError( Error ):
    def __init__( self, desc, start_pos, end_pos=None ):
        super().__init__( 'Syntax Error', desc, start_pos, end_pos )


class NameNotFoundError( Error ):
    def __init__( self, desc, start_pos, end_pos=None ):
        super().__init__( 'Name Error', desc, start_pos, end_pos )


#!/usr/bin/env python3

import sys 
import errno

from cicinlang.cicinlang import run 

def main():
    if len( sys.argv ) != 2:
        print( "Usage: cicinlang <path_to_file>" )
        return 

    path = sys.argv[1]

    try:
        with open( path, "r" ) as f:
            statement = f.read()
        
        try: 
            _, error = run( '<stdin>', statement )

        except KeyboardInterrupt:
            pass 

        else:
            if error: print( error )
    
    except IOError as x:
        if x.errno == errno.ENOENT:
            print( f"Error: {path} does not exist" )
        
        elif x.errno == errno.EISDIR:
            print( f"Error: {path} is a directory" )

        elif x.errno == errno.EACCES:
            print( f"Error: You don't have permissions to open {path}" )

        else:
            print( "Unknown Error. Please try again." )
            return 
    
    except: 
        print( "Unknown Error. Please try again." )
        return 

if __name__ == '__main__':
    main()


import string

class Constants:
    LETTERS = string.ascii_letters
    DIGITS = '0123456789'
    OPERATORS = '+-*/^'
    WHITESPACES = ' \t\n'
    PARENTHESES = '()'
    KEYWORDS = { 'var', 'and', 'or', 'not', 'if', 'elif', 'else', 'for', 'input_str', 'input_num', 'print', 'str' }

    AT_OLD = 'OLD'
    AT_NEW = 'NEW'

    TT_INT = 'INT'
    TT_FLOAT = 'FLOAT'
    TT_STRING = 'STRING'

    TT_PLUS = 'PLUS'
    TT_MINUS = 'MINUS'
    TT_MUL = 'MUL'
    TT_DIV = 'DIV'
    TT_EXP = 'EXP'
    TT_LPAREN = 'LPAREN'
    TT_RPAREN = 'RPAREN'
    TT_LBRACE = 'LBRACE'
    TT_RBRACE = 'RBRACE'
    TT_EOF = 'EOF'
    TT_EQ = 'EQ'
    TT_EE = 'EE'
    TT_NE = 'NE'
    TT_GT = 'GT'
    TT_GTE = 'GTE'
    TT_LT = 'LT'
    TT_LTE = 'LTE'
    TT_IDENTIFIER = 'IDENTIFIER'
    TT_KEYWORD = 'KEYWORD'
    TT_SEMICOLON = 'SEMICOLON'
    TT_COMMA = 'COMMA'


class Token:
    def __init__( self, type_, value=None ):
        self.type = type_
        self.value = value 
    
    def __repr__( self ):
        if self.value != None: return f'{self.type}:{self.value}'
        else: return f'{self.type}'

    def set_pos( self, start_pos, end_pos=None ):
        self.start_pos = start_pos 

        if end_pos == None:
            pos = start_pos.copy()
            pos.advance()
            self.end_pos = pos 

        else: 
            self.end_pos = end_pos 
        
        return self 
    

class Position:
    def __init__( self, fn, ftext, idx=-1, ln=0, coln=-1 ):
        self.fn = fn 
        self.ftext = ftext 
        self.idx = idx
        self.ln = ln
        self.coln = coln 

        if self.idx == -1:
            self.advance()
    
    def advance( self ):
        self.idx += 1
        self.coln += 1

        if self.idx > 0 and self.idx <= len( self.ftext ) and self.ftext[self.idx - 1] == '\n':
            self.ln += 1
            self.coln = 0
    
    def copy( self ):
        return Position( self.fn, self.ftext, self.idx, self.ln, self.coln )
    

class Context:
    def __init__( self, name, parent=None, parent_entry_pos=None ):
        self.name = name 
        self.parent = parent 
        self.parent_entry_pos = parent_entry_pos 
        self.symbol_table = None 
    
    def set_symbol_table( self, symbol_table ):
        self.symbol_table = symbol_table


class SymbolTable:
    def __init__( self, context ):
        self.table = {}
        self.context = context 
    
    def get_value( self, key ):
        val = self.table.get( key, None )

        if val == None and self.context.parent != None:
            return self.context.parent.table.get( key, None )

        return val 

    def set_value( self, key, value ):
        self.table[key] = value 

    def remove_value( self, key ):
        del self.table[key]



from cicinlang.utils import Token, Position, Constants
from cicinlang.errors import IllegalCharException

class Lexer:
    def __init__( self, fn, ftext ):
        self.pos = Position( fn, ftext )
        self.cur_char = ftext[self.pos.idx] if self.pos.idx < len( ftext ) else None
    
    def advance( self ):
        self.pos.advance()
        self.cur_char = self.pos.ftext[self.pos.idx] if self.pos.idx < len( self.pos.ftext ) else None

    def create_tokens( self ):
        tokens = []

        while self.cur_char != None:
            if self.cur_char in Constants.WHITESPACES:
                self.advance()

            elif self.cur_char in Constants.LETTERS:
                tokens.append( self.create_identifier_token() )

            elif self.cur_char in Constants.DIGITS:
                tokens.append( self.create_num_token() )
            
            elif self.cur_char in Constants.OPERATORS:
                tokens.append( self.create_operator_token() )
            
            elif self.cur_char in Constants.PARENTHESES:
                tokens.append( self.create_parentheses_token() )

            elif self.cur_char == '>':
                tokens.append( self.create_greater_than_token() )
            
            elif self.cur_char == '<':
                tokens.append( self.create_less_than_token() )
            
            elif self.cur_char == '=':
                tokens.append( self.create_equals_token() )

            elif self.cur_char == '!':
                token, error = self.create_not_equals_token()

                if error: return None, error 

                tokens.append( token )

            elif self.cur_char == '#':
                while self.cur_char not in  [ '\n', None ]:
                    self.advance()
            
            elif self.cur_char == ';':
                tokens.append( Token( Constants.TT_SEMICOLON ).set_pos( self.pos ) )
                self.advance()

            elif self.cur_char == ',':
                tokens.append( Token( Constants.TT_COMMA ).set_pos( self.pos ) )
                self.advance()

            elif self.cur_char == '{':
                tokens.append( Token( Constants.TT_LBRACE ).set_pos( self.pos ) )
                self.advance()

            elif self.cur_char == '}':
                tokens.append( Token( Constants.TT_RBRACE ).set_pos( self.pos ) )
                self.advance()

            elif self.cur_char in [ "'", '"']:
                tok, error = self.create_string_token()

                if error: return None, error 

                tokens.append( tok )

            else:
                c = "'" + self.cur_char + "'"
                return None, IllegalCharException( c, self.pos )
        
        tokens.append( Token( Constants.TT_EOF ).set_pos( self.pos, self.pos ) )
        return tokens, None 
    
    def create_greater_than_token( self ):
        start_pos = self.pos.copy() 
        tok_type = Constants.TT_GT 
        tok_val = '>'
        self.advance()

        if self.cur_char == '=':
            tok_type = Constants.TT_GTE
            tok_val = '>='
            self.advance()
        
        return Token( tok_type, tok_val ).set_pos( start_pos, self.pos )
        
    def create_less_than_token( self ):
        start_pos = self.pos.copy() 
        tok_type = Constants.TT_LT 
        tok_val = '<'
        self.advance()

        if self.cur_char == '=':
            tok_type = Constants.TT_LTE
            tok_val = '<='
            self.advance()
        
        return Token( tok_type, tok_val ).set_pos( start_pos, self.pos )
    
    def create_equals_token( self ):
        start_pos = self.pos.copy() 
        tok_type = Constants.TT_EQ
        tok_val = None
        self.advance()

        if self.cur_char == '=':
            tok_type = Constants.TT_EE 
            tok_val = '=='
            self.advance()
        
        return Token( tok_type, tok_val ).set_pos( start_pos, self.pos )
    
    def create_not_equals_token( self ):
        start_pos = self.pos.copy() 
        self.advance()
        c = self.cur_char
        self.advance()

        if self.cur_char == '=':
            tok_type = Constants.TT_NE
            tok_val = '!='
            return Token( tok_type ).set_pos( start_pos, self.pos ), None 

        else:
            return None, IllegalCharException( f"'{c}'", start_pos, self.pos )
    
    
    def create_identifier_token( self ):
        id_str = ''
        pos_start = self.pos.copy()
        
        while self.cur_char != None and self.cur_char in Constants.LETTERS + Constants.DIGITS + '_':
            id_str += self.cur_char 
            self.advance()
        
        tok_type = Constants.TT_KEYWORD if id_str in Constants.KEYWORDS else Constants.TT_IDENTIFIER

        return Token( tok_type, id_str ).set_pos( pos_start, self.pos )

    def create_num_token( self ):
        num_str = ''
        dot_count = 0
        start_pos = self.pos.copy()

        while self.cur_char != None and self.cur_char in Constants.DIGITS + '.':
            if self.cur_char == '.':
                if dot_count >= 1: break 
                else: dot_count += 1

            num_str += self.cur_char 
            self.advance()

        if dot_count > 0: return Token( Constants.TT_FLOAT, float( num_str ) ).set_pos( start_pos, self.pos )
        else: return Token( Constants.TT_INT, int( num_str ) ).set_pos( start_pos, self.pos )

    def create_operator_token( self ):
        if self.cur_char == '+':
            type_ = Constants.TT_PLUS
            val = '+'

        elif self.cur_char == '-':
            type_ = Constants.TT_MINUS
            val = '-'

        elif self.cur_char == '*':
            type_ = Constants.TT_MUL
            val = '*'

        elif self.cur_char == '/':
            type_ = Constants.TT_DIV
            val = '/'

        elif self.cur_char == '^':
            type_ = Constants.TT_EXP
            val = '^'
        
        tok = Token( type_, val ).set_pos( self.pos )
        self.advance()
        return tok 

    def create_parentheses_token( self ):
        if self.cur_char == '(':
            type_ = Constants.TT_LPAREN

        elif self.cur_char == ')':
            type_ = Constants.TT_RPAREN
        
        tok = Token( type_ ).set_pos( self.pos )
        self.advance()
        return tok 
    
    def create_string_token( self ):
        end_char = self.cur_char 
        start_pos = self.pos.copy()
        s = ""

        self.advance()

        while self.cur_char != None and self.cur_char != end_char:
            s += self.cur_char
            self.advance()
        
        if self.cur_char == None:
            if end_char == '"':
                char_s = "'\"'"
            
            else:
                char_s = '"\'"'

            return None, IllegalCharException( f'{char_s} expected', self.pos )

        self.advance()

        return Token( Constants.TT_STRING, s ).set_pos( start_pos, self.pos ), None




from cicinlang.lexer import Lexer 
from cicinlang.parser_ import Parser
from cicinlang.interpreter import Interpreter
from cicinlang.utils import Context, SymbolTable

global_context = Context( '<module>', None, None )
global_symbol_table = SymbolTable( global_context )
global_context.set_symbol_table( global_symbol_table )

def run( fn, ftext ):
    lexer = Lexer( fn, ftext )
    tokens, error = lexer.create_tokens()

    if error: return None, error 

    if len( tokens ) == 1: return '', None 

    parser = Parser( tokens )
    ast_list, error = parser.parse()

    if error: return None, error 

    interpreter = Interpreter()
    res, error = interpreter.interpret( ast_list, global_context )

    if error: return None, error

    return res, None


from cicinlang.utils import Constants, Context, SymbolTable
from cicinlang.errors import RuntimeException, NameNotFoundError

class Number:
    def __init__( self, value ):
        self.value = value 


class String:
    def __init__( self, value ):
        self.value = value 


class Function:
    def __init__( self, args, nodes ):
        self.args = args 
        self.nodes = nodes

class ReturnValue:
    def __init__( self, payload ):
        self.payload = payload


class NoneValue:
    def __init__( self ):
        self.value = None 


class Interpreter:   
    def interpret( self, ast_list, context ):
        for ast in ast_list:
            res, error = self.visit( ast, context ) 

            if error: return None, error

            if type( res ).__name__ == 'ReturnValue':
                return res, None 
        
        return NoneValue(), None 

    def visit( self, root, context ):
        type_name = type( root ).__name__ 

        if type_name == 'BinOpNode':
            left, error = self.visit( root.left_node, context )

            if error: return None, error 

            l_type = type( left ).__name__

            if l_type == 'ReturnValue':
                left = left.payload

            tok = root.tok

            right, error = self.visit( root.right_node, context )

            if error: return None, error

            r_type = type( right ).__name__

            if r_type == 'ReturnValue':
                right = right.payload

            is_func = l_type == 'Function' or r_type == 'Function'

            diff_types_wo_and_or_ee_ne = ( l_type != r_type ) and ( tok.type not in [ Constants.TT_KEYWORD, Constants.TT_EE, Constants.TT_NE ] )

            str_and_not_add_and_or = l_type == 'String' and tok.type not in [ Constants.TT_PLUS, Constants.TT_EE, Constants.TT_NE, Constants.TT_KEYWORD ]

            if is_func or diff_types_wo_and_or_ee_ne or str_and_not_add_and_or:
                return None, RuntimeException( f"Unsupported binary operation '{tok.value}' on types: '{l_type}' and '{r_type}'", root.start_pos, root.end_pos )
            
            return self.bin_op( left, tok, right )

        elif type_name == 'UnOpNode':
            res, error = self.interpret( root.node, context )

            if error: return None, error 

            res_type = type( res ).__name__

            if res_type == 'ReturnValue':
                res = res.payload

            is_func = res_type == 'Function'

            str_and_not_not = res_type == 'String' and root.tok.value != 'not'

            if is_func or str_and_not_not:
                return None, RuntimeException( f"Unsupported unary operator '{root.tok.value}' on type '{res_type}'", root.start_pos, root.end_pos )

            tok = root.tok 

            if tok.type == Constants.TT_PLUS: return Number( res.value ), None 

            elif tok.type == Constants.TT_KEYWORD and tok.value == 'not': return ( 0 if res.value else 1 ), None 

            else: return Number( -1 * num.value ), None 
        
        elif type_name == 'NumberNode':
            return Number( root.tok.value ), None 
        
        elif type_name == 'StringNode':
            return String( root.tok.value ), None 

        elif type_name == 'VarAssignNode':
            val, error = self.visit( root.var_value_node, context )

            if error: return None, error 

            if type( val ).__name__ == 'ReturnValue':
                val = val.payload

            tok = root.var_name_token.value
            tok_val = context.symbol_table.get_value( tok )
            start_pos = root.var_name_token.start_pos
            end_pos = root.var_name_token.end_pos

            if root.assign_type == Constants.AT_OLD and tok_val == None:
                return None, NameNotFoundError( f"'{tok}' not found", start_pos, end_pos )

            elif root.assign_type == Constants.AT_NEW and tok_val != None:
                return None, NameNotFoundError( f"'{tok}' is already declared.", start_pos, end_pos )

            context.symbol_table.set_value( tok, val )

            return val, None 

        elif type_name == 'VarAccessNode':
            value = context.symbol_table.get_value( root.var_name_token.value )
            
            if value == None:
                return None, NameNotFoundError( f"'{root.var_name_token.value}' not found", root.var_name_token.start_pos, root.var_name_token.end_pos )

            return value, None 

        elif type_name == 'IfNode':
            res, error = self.visit( root.if_condition_node, context )

            if error: return None, error 

            if type( res ).__name__ == 'ReturnValue':
                res = res.payload

            if type( res ).__name__ != 'Number':
                return None, RuntimeException( f"Condition value for if statement can only be Number, got {type( res ).__name__}", root.if_condition_node.start_pos, root.if_condition_node.end_pos )
            
            cond_val = res.value 

            if cond_val == 1:
                res, error = self.interpret( root.if_body_node_list, context )

                if error: return None, error 
                
                return res, None 

            for elif_cond_node, elif_body_node_list in zip( root.elif_condition_node_list, root.elif_body_node_lists ):
                res, error = self.visit( elif_cond_node )

                if error: return None, error 

                if type( res ).__name__ == 'ReturnValue':
                    res = res.payload

                if type( res ).__name__ != 'Number':
                    return None, RuntimeException( f"Condition value for if statement can only be Number, got {type( res ).__name__}", root.if_condition_node.start_pos, root.if_condition_node.end_pos )
                
                cond_val = res.value 

                if cond_val == 1:
                    res, error = self.interpret( elif_body_node_list, context )

                    if error: return None, error 

                    return res, None
            
            res, error = self.interpret( root.else_body_node_list, context )

            if error: return None, error 
            
            return res, None
          
        elif type_name == 'ForNode':
            if root.init_node != None:
                _, error = self.visit( root.init_node, context )

                if error: return None, error 

            cond_res, error = self.visit( root.cond_node, context )

            if error: return None, error 

            if type( cond_res ).__name__ == 'ReturnValue':
                cond_res = cond_res.payload

            cond_res_type = type( cond_res ).__name__

            if cond_res_type != "Number":
                return None, RuntimeException( f"Output of condition statement in for loop should be number, got {cond_res_type}", root.cond_node.start_pos, root.cond_node.end_pos )

            cond_val = cond_res.value

            while cond_val == 1:
                res, error = self.interpret( root.body_node_list, context )

                if error: return error 

                if type( res ).__name__ == 'ReturnValue':
                    return res, None 
                
                _, error = self.visit( root.update_node, context )

                if error: return None, error 

                cond_res, error = self.visit( root.cond_node, context )

                if error: return None, error 

                if type( cond_res ).__name__ == 'ReturnValue':
                    cond_res = cond_res.payload

                cond_res_type = type( cond_res ).__name__

                if cond_res_type != "Number":
                    return None, RuntimeException( f"Output of condition statement in for loop should be number, got {cond_res_type}", root.cond_node.start_pos, root.cond_node.end_pos )

                cond_val = cond_res.value
            
            return NoneValue(), None 

        elif type_name == 'FunctionDefNode':
            return Function( root.func_arg_toks_list, root.func_body_nodes_list ), None 

        elif type_name == 'FunctionCallNode':
            func_name = root.func_name_tok.value

            func = context.symbol_table.get_value( func_name )

            if func == None:
                return None, NameNotFoundError( f"{func_name} not found", root.func_name_tok.start_pos, root.func_name_tok.end_pos )
            
            if type( func ).__name__ != 'Function':
                return None, RuntimeException( f"Cannot invoke function call on type { type( func ).__name__ }", root.func_name_tok.start_pos, root.func_name_tok.end_pos )

            child_context = Context( func_name, context, root.func_name_tok.start_pos )
            child_symbol_table = SymbolTable( child_context )
            child_context.set_symbol_table( child_symbol_table )

            n_params = len( func.args )
            n_args = len( root.func_arg_nodes_list )

            if n_params != n_args:
                return None, RuntimeException( f"Expected {n_params} arguments, got {n_args}", root.start_pos, root.end_pos )

            for param_tok, node in zip( func.args, root.func_arg_nodes_list ):
                param_name = param_tok.value
                res, error = self.visit( node, context )

                if error: return None, error 

                if type( res ).__name__ == 'ReturnValue':
                    res = res.payload

                child_context.symbol_table.set_value( param_name, res )
            
            res, error = self.interpret( func.nodes, child_context )

            if error: return None, error 

            return res, None 

        elif type_name == 'PrintNode':
            res, error = self.visit( root.node, context )

            if error: return None, error 

            if type( res ).__name__ == 'ReturnValue':
                res = res.payload

            if type( res ).__name__ == 'Function':
                return None, RuntimeException( "Cannot print functions", root.node.start_pos, root.node.end_pos )

            print( res.value )

            return NoneValue(), None 
        
        elif type_name == 'InputNode':
            if root.node != None:
                res, error = self.visit( root.node, context )

                if error: return None, error 

                if type( res ).__name__ == 'ReturnValue':
                    res = res.payload

                if type( res ).__name__ == 'Function':
                    return None, RuntimeException( "Cannot print functions", root.node.start_pos, root.node.end_pos )
            
            else:
                res = None

            inp = input( res.value if res != None else '' )

            if root.inp_type == 'str':
                return String( inp ), None 
            
            elif root.inp_type == 'num':
                try: 
                    num = float( inp ) if '.' in inp else int( inp )

                except:
                    return None, RuntimeException( f"Could not convert '{inp}' to Number", root.start_pos, root.end_pos )

                else:
                    return Number( num ), None 

        elif type_name == 'ReturnNode':
            res, error = self.visit( root.node, context )

            if error: return None, error 

            if type( res ).__name__ == 'ReturnValue':
                return res, None 
            
            return ReturnValue( res ), None 

        elif type_name == 'StringifyNode':
            res, error = self.visit( root.node, context )

            if error: return None, error 

            if type( res ).__name__ == 'ReturnValue':
                res = res.payload
            
            if type( res ).__name__ == 'Function':
                return String( f'<Function object>' ), None 

            else: 
                return String( str( res.value ) ), None

        elif type_name == 'NoneType':
            pass

        else:
            return None, 'Interpreting Error'
    
    def bin_op( self, left, tok, right ):
        if tok.type == Constants.TT_PLUS: 
            if type( left ).__name__ == 'Number':
                return Number( left.value + right.value ), None 
            
            elif type( left ).__name__ == 'String':
                return String( left.value + right.value ), None 
            
        elif tok.type == Constants.TT_MINUS: return Number( left.value - right.value ), None 

        elif tok.type == Constants.TT_MUL: return Number( left.value * right.value ), None 

        elif tok.type == Constants.TT_EXP: return Number( left.value ** right.value ), None

        elif tok.type == Constants.TT_DIV:
            if right.value == 0: return None, RuntimeException( 'Division by 0', tok.start_pos )

            return Number( left.value / right.value ), None

        elif tok.type == Constants.TT_LT: return Number( 1 if left.value < right.value else 0 ), None 

        elif tok.type == Constants.TT_LTE: return Number( 1 if left.value <= right.value else 0 ), None 

        elif tok.type == Constants.TT_GT: return Number( 1 if left.value > right.value else 0 ), None

        elif tok.type == Constants.TT_GTE: return Number( 1 if left.value >= right.value else 0 ), None

        elif tok.type == Constants.TT_EE: return Number( 1 if left.value == right.value else 0 ), None

        elif tok.type == Constants.TT_NE: return Number( 1 if left.value != right.value else 0 ), None

        elif tok.type == Constants.TT_KEYWORD and tok.value == 'and': return Number( 1 if left.value and right.value else 0 ), None

        elif tok.type == Constants.TT_KEYWORD and tok.value == 'or': return Number( 1 if left.value or right.value else 0 ), None


class NumberNode:
    def __init__( self, tok ):
        self.tok = tok 
        self.start_pos = tok.start_pos
        self.end_pos = tok.end_pos

    def __repr__( self ):
        return f'{self.tok}'
    

class StringNode:
    def __init__( self, tok ):
        self.tok = tok 
        self.start_pos = tok.start_pos
        self.end_pos = tok.end_pos

    def __repr__( self ):
        return f'{self.tok}'


class BinOpNode:
    def __init__( self, left_node, tok, right_node ):
        self.left_node = left_node
        self.tok = tok
        self.right_node = right_node
        self.start_pos = left_node.start_pos
        self.end_pos = right_node.end_pos

    def __repr__( self ):
        return f'( {self.left_node}, {self.tok}, {self.right_node} )'


class UnOpNode:
    def __init__( self, tok, node ):
        self.tok = tok 
        self.node = node 
        self.start_pos = tok.start_pos 
        self.end_pos = node.end_pos
    
    def __repr__( self ):
        return f'( {self.tok}, {self.node} )'
    

class VarAssignNode:
    def __init__( self, var_name_token, var_value_node, assign_type ):
        self.var_name_token = var_name_token  # This token is an IDENTIFIER token
        self.var_value_node = var_value_node  # This node will be an expression node
        self.start_pos = var_name_token.start_pos
        self.end_pos = var_value_node.end_pos
        self.assign_type = assign_type
    
    def __repr__( self ):
        return f'( {self.var_name_token.value} : {self.var_value_node} )'


class VarAccessNode:
    def __init__( self, var_name_token ):
        self.var_name_token = var_name_token 
        self.start_pos = var_name_token.start_pos
        self.end_pos = var_name_token.end_pos

    def __repr__( self ):
        return f'{self.var_name_token.value}'
    

class IfNode:
    def __init__( self, if_condition_node, if_body_node_list, elif_condition_node_list, elif_body_node_lists, else_body_node_list ):
        self.if_condition_node = if_condition_node
        self.if_body_node_list = if_body_node_list
        self.elif_condition_node_list = elif_condition_node_list
        self.elif_body_node_lists = elif_body_node_lists
        self.else_body_node_list = else_body_node_list
        self.start_pos = if_condition_node.start_pos if if_condition_node != None else None
        
        if len( else_body_node_list ) > 0:
            self.end_pos = else_body_node_list[-1].end_pos

        elif len( elif_condition_node_list ) > 0:
            if len( elif_body_node_lists[-1] ) > 0:
                self.end_pos = elif_body_node_lists[-1][-1].end_pos
                
            else:
                self.end_pos = elif_condition_node_list[-1].end_pos

        elif len( if_body_node_list ) > 0:
            self.end_pos = if_body_node_list[-1].end_pos
        
        else:
            self.end_pos = if_condition_node.end_pos if if_condition_node != None else None

    def __repr__( self ):
        return f'( IF: {self.if_condition_node}, {self.if_body_node_list}, ELIF: {self.elif_condition_node_list}, {self.elif_body_node_lists}, ELSE: {self.else_body_node_list} )'
    

class ForNode:
    def __init__( self, init_node, cond_node, update_node, body_node_list ):
        self.init_node = init_node  
        self.cond_node = cond_node 
        self.update_node = update_node 
        self.body_node_list = body_node_list
        self.start_pos = init_node.start_pos if init_node != None else cond_node.start_pos
        self.end_pos = body_node_list[-1].end_pos if len( body_node_list ) > 0 else update_node.end_pos

    def __repr__( self ):
        return f'( FOR: {self.init_node}, {self.cond_node}, {self.update_node}, {self.body_node_list} )'


class FunctionDefNode:
    def __init__( self, func_arg_toks_list, func_body_nodes_list ):
        self.func_arg_toks_list = func_arg_toks_list
        self.func_body_nodes_list = func_body_nodes_list
        self.start_pos = func_arg_toks_list[0].start_pos if len( func_arg_toks_list ) > 0 else func_body_nodes_list[0].start_pos
        self.end_pos = func_body_nodes_list[-1].end_pos
    
    def __repr__( self ):
        return f'( FUNC - ARGS:{self.func_arg_toks_list}, BODY:{self.func_body_nodes_list} )'


class FunctionCallNode:
    def __init__( self, func_name_tok, func_arg_nodes_list ):
        self.func_name_tok = func_name_tok
        self.func_arg_nodes_list = func_arg_nodes_list
        self.start_pos = func_name_tok.start_pos
        self.end_pos = func_arg_nodes_list[-1].end_pos if len( func_arg_nodes_list ) > 0 else func_name_tok.end_pos

    def __repr__( self ):
        return f'( CALLING FUNC "{self.func_name_tok.value}" WITH ARGS {self.func_arg_nodes_list} )'


class ReturnNode:
    def __init__( self, val_node ):
        self.node = val_node
        self.start_pos = val_node.start_pos
        self.end_pos = val_node.end_pos
    
    def __repr__( self ):
        return f'( RETURN {self.node} )'


class PrintNode:
    def __init__( self, node ):
        self.node = node 
        self.start_pos = node.start_pos
        self.end_pos = node.end_pos
    
    def __repr__( self ):
        return f'( PRINT: {self.node} )'
    

class InputNode:
    def __init__( self, node, inp_type ):
        self.node = node 
        self.inp_type = inp_type
        self.start_pos = node.start_pos if node != None else None
        self.end_pos = node.end_pos if node != None else None 
    
    def __repr__( self ):
        return f'( INPUT({self.inp_type}) - MSG: {self.node} )'


class StringifyNode:
    def __init__( self, node ):
        self.node = node
        self.start_pos = node.start_pos 
        self.end_pos = node.end_pos